#include "config.h"
#include "defines.h"
#include "fdbuf.h"
#include <dirent.h>
#include <errno.h>
#include <iostream.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include "list.h"

typedef list<mystring> slist;
typedef list_iterator<mystring> slist_iterator;

#define fail(MSG) do { cerr << MSG << endl; return false; } while(0)

static slist remotes;
static bool reload_remotes = false;

void catch_hup(int)
{
  signal(SIGHUP, catch_hup);
  reload_remotes = true;
}

bool load_remotes()
{
  remotes.empty();
  mystring remote;
  fdibuf in(CONFIG_DIR "remotes");
  while(in.getline(remote))
    remotes.append(remote);
  return remotes.count() > 0;
}

static slist files;
static bool reload_files = false;

void catch_alrm(int)
{
  signal(SIGALRM, catch_alrm);
  reload_files = true;
}

bool load_files()
{
  DIR* dir = opendir(".");
  if(!dir)
    fail("Cannot open queue directory: " << strerror(errno));
  files.empty();
  struct dirent* entry;
  while((entry = readdir(dir)) != 0) {
    const char* name = entry->d_name;
    if(name[0] == '.')
      continue;
    files.append(name);
  }
  closedir(dir);
  return true;
}

bool sethostbyname(mystring hostname, struct sockaddr_in& sa)
{
  struct hostent *he = gethostbyname(hostname.c_str());
  if(!he) {
    switch(h_errno) {
    case HOST_NOT_FOUND:
      cout << "Host not found." << endl;
      break;
    case NO_ADDRESS:
      cout << "Host has no address records." << endl;
      break;
    case NO_RECOVERY:
      cout << "Unrecoverable error in gethostbyaddr." << endl;
      break;
    case TRY_AGAIN:
      cout << "Temporary error in gethostbyaddr." << endl;
      break;
    default:
      cout << "Unknown error in gethostbyaddr." << endl;
    }
    return false;
  }
  memcpy(&sa.sin_addr, he->h_addr, he->h_length);
  return true;
}

int connect(mystring hostname, int port)
{
  cout << "Connecting to host '" << hostname << "' on port " << port << endl;
  struct sockaddr_in sa;
  memset(&sa, 0, sizeof(sa));
  if(!sethostbyname(hostname, sa))
    return -1;
  sa.sin_port = htons(port);
  int s = socket(AF_INET, SOCK_STREAM, 0);
  if(s == -1) {
    cout << "Call to socket failed: " << strerror(errno) << endl;
    return -1;
  }
  if(connect(s, (const sockaddr*)&sa, sizeof(sa)) != 0) {
    cout << "Connect failed: " << strerror(errno) << endl;
    return -1;
  }
  return s;
}

bool exec_protocol(mystring filename, int fd, mystring protocol)
{
  if(close(0) == -1 || close(1) == -1)
    fail("Error closing file descriptors.");
  if(dup2(fd, 0) == -1 || dup2(fd, 1) == -1)
    fail("Error duplicating the socket.");
  if(close(fd) == -1)
    fail("Error closing the old socket descriptor.");
  mystring program(PROTOCOL_DIR, protocol.c_str());
  execl(program.c_str(), program.c_str(), filename.c_str(), 0);
  fail("Could not exec '" << program << "'");
}

#undef fail
#define fail(MSG) do { cout << MSG << endl; return false; } while(0)

bool catchsender(pid_t pid)
{
  int status;
  if(waitpid(pid, &status, 0) == -1)
    fail("Error catching the child process return value.");
  else {
    if(WIFEXITED(status)) {
      status = WEXITSTATUS(status);
      if(status)
	fail("Sending failed.");
      else {
	cout << "Sent file." << endl;
	return true;
      }
    }
    else
      fail("Sending process crashed or was killed.");
  }
  return false;
}

bool send_one(mystring filename, mystring remote, mystring proto, int port)
{
  int fd = connect(remote, port);
  if(fd < 0)
    return false;
  pid_t pid = fork();
  switch(pid) {
  case -1:
    fail("Fork failed.");
  case 0:
    exec_protocol(filename, fd, proto);
    exit(111);
    return false;
  default:
    close(fd);
    if(!catchsender(pid))
      return false;
    if(unlink(filename.c_str()) == -1)
      fail("Can't unlink file.");
    return true;
  }
}

bool send_all()
{
  if(remotes.count() <= 0)
    fail("No remote hosts listed for delivery");
  if(files.count() == 0)
    return true;
  cout << "Starting delivery, " << files.count() << " message(s) in queue."
       << endl;
  for(slist_iterator remote(remotes); remote; remote++) {
    for(slist_iterator file(files); file; file++) {
      if(!send_one(*file, *remote, "smtp", 25))
	break;
      files.remove(*file);
    }
  }
  cout << "Delivery complete, " << files.count() << " message(s) remain."
       << endl;
  return true;
}

static int trigger;

bool open_trigger()
{
  trigger = open(QUEUE_TRIGGER, O_RDONLY|O_NONBLOCK);
  if(trigger == -1)
    fail("Could not open trigger file.");
  return true;
}

bool read_trigger()
{
  if(trigger != -1)
    close(trigger);
  return open_trigger();
}

bool do_select()
{
  fd_set readfds;
  FD_ZERO(&readfds);
  FD_SET(trigger, &readfds);
  struct timeval timeout;
  timeout.tv_sec = 60;
  timeout.tv_usec = 0;
  int s = select(trigger+1, &readfds, 0, 0,
		 (files.count() == 0) ? 0 : &timeout);
  if(s == 1) {
    cout << "Trigger pulled, rescanning queue." << endl;
    read_trigger();
    reload_files = 1;
  }
  else if(s == -1 && errno != EINTR)
    fail("Internal error in select.");
  if(reload_remotes) {
    cout << "Signal received, rescanning control files." << endl;
    load_remotes();
  }
  if(reload_files) {
    cout << "Rescanning queue." << endl;
    load_files();
  }
  return true;
}

int main(int, char*[])
{
  if(!open_trigger())
    return 1;
  if(chdir(QUEUE_MSG_DIR) == -1) {
    cout << "Could not chdir to queue message directory." << endl;
    return 1;
  }
  
  signal(SIGALRM, catch_alrm);
  signal(SIGHUP, catch_hup);
  load_remotes();
  load_files();
  for(;;) {
    send_all();
    do_select();
  }
  return 0;
}
